// Or.java

import swarm.Globals;
import swarm.collections.ListImpl;

/**
 * An Or manager for our orders<br><br>
 *
 * orCriterion<br>
 *			0 - all 'or' branches in sequence<br>
 *			1 - choosing first branch<br>
 *			2 - choosing second branch<br>
 *			3 - a random branch<br>
 *			4 - the branch whose first step
 *			    has the shortest waiting list<br>
 *                      5 - the branch whose number is stored in (x,0) pos in
 *                          memoryMatrix designated by orMemoryMatrix in
 *                          VEFrameModelSwarm.java; the matrix 
 *                          may be sensitive or insensitive to layers;
 *                          range of the branch number: from 1 to 
 *                          branchNumber <br>
 *                          x is 0 if the first node in 'or' sequence is
 *                          numbered 1; is kk if the first node is numbered
 *                          10kk with kk 00 to 99<br><br>
 *
 * @author Pietro Terna
 */
public class Or
{
    // the class is used in a static way, so we have no construtors 
    // here and we have to declare static all the methods

    /** or criterion to be applied */
    public static int orCriterion;
    /** the list of the operating units */
    public static ListImpl unitList;
    /** memory matrix addresses to be used in orCriterion 5 */
    public static MemoryMatrix mm;


    /** setting the 'or' criterion */
    public static void setOrCriterion(int k){
	orCriterion=k;
	if(orCriterion<0 || orCriterion > 6){
	    System.out.println("Invalid OrCriterion value.");
	    MyExit.exit(1);
	}
    }

    /** setting the list of the units */
    public static void setUnitList(ListImpl ul){
	unitList=ul;
    }


    /** setting the address of memory matrix */
    public static void setMemoryMatrix(MemoryMatrix m){
	mm=m;
	}


    /** applying 'or' */
    public static boolean applyOr(Order o){
	OrSpecificationSet spec; 
	Unit aUnit;
	boolean unitNotFound;
	int i, iii, node, chosenNode, length, length0, tmp;
	int [] recipe, state;

	// the or specification set to be used
	spec=(OrSpecificationSet) o.getPendingOrSpecificationSet();
	if(spec == null){
	    System.out.println("Internal error: found an order x in " +
			       "Or.applyOr(x) without 'or' specifications");
	    MyExit.exit(1);
	}

	// applying the 'or' criterion
	if(orCriterion == 0)return true;

	chosenNode=0;
	recipe = (int []) o.getRecipeVector();
	state =  (int []) o.getStateVector();

	if(orCriterion == 1)chosenNode=1;
	if(orCriterion == 2)chosenNode=2;
	if(orCriterion == 3)chosenNode=Globals.env.uniformIntRand.
			     getIntegerWithMin$withMax(1,spec.getNodeNumber());

	if(orCriterion == 4)
	    {
		length=1000000000;
		// looking for the first unit in each branch
		for (node=1; node<=spec.getNodeNumber();node++)
		    {
			length0=length;
			unitNotFound=true;
			for (i=0;i<unitList.getCount() && unitNotFound;i++)
			    {
				aUnit=(Unit) unitList.atOffset(i);
				for (iii=0;iii<aUnit.getNOfPhasesToDealWith() 
					 && unitNotFound;iii++)
				    {
					if(recipe[spec.getNodePosition(node)]==
					   aUnit.getProductionPhase(iii))
					    {
						unitNotFound=false;
						length=aUnit.
						    getWaitingListLength();
						if(length<length0)chosenNode=
								      node;
					    }
				    }
			    }
			
		    }
	    }

	if(orCriterion == 5)
	    {
		if(mm.getEmpty(
		  o.getOrderLayer(),
		  ((OrSpecificationSet) o.getPendingOrSpecificationSet()).
		    getOrRow(),
		  0))
		    return false;

		tmp=(int) mm.getValue(
		  o.getOrderLayer(),
		  ((OrSpecificationSet) o.getPendingOrSpecificationSet()).
		    getOrRow(),
		  0);
		
		if(tmp < 1 || tmp > spec.getNodeNumber()){
		    System.out.println("Invalid branch number in Or process"+
				       " in order # " + o.getOrderNumber());
		    MyExit.exit(1);
		}
		
		chosenNode=tmp;
	    }



	if(StartVEFrame.verbose)
	    System.out.println("Applying Or to order number "+
			       o.getOrderNumber()+
			       " at pos. "+spec.getPositionInRecipe());

	for (node=1; node<=spec.getNodeNumber();node++)
	    {
		for(i=spec.getNodePosition(node);
		    i<spec.getNodePosition(node+1);i++)
		    {
			if(chosenNode != node) state[i]=-1;
		    }
	    }

	return true;

    }

}
